import numpy as np
from functools import partial
from typing import *
from .llm_model import (
    get_llm,
    openai_choice_predict,
    openai_topk_predict,
    DiscreteDist,
    GaussDist,
    wrap_chatllm,
)
from .aqfxns import (
    probability_of_improvement,
    expected_improvement,
    upper_confidence_bound,
    greedy,
)
from .pool import Pool
from langchain.prompts.few_shot import FewShotPromptTemplate
from langchain.prompts.prompt import PromptTemplate
from langchain.vectorstores import FAISS, Chroma
from langchain.embeddings import OpenAIEmbeddings
from langchain.prompts.example_selector import (
    MaxMarginalRelevanceExampleSelector,
    SemanticSimilarityExampleSelector,
)

_answer_choices = ["A", "B", "C", "D", "E"]


class QuantileTransformer:
    def __init__(self, values, n_quantiles):
        self.n_quantiles = n_quantiles
        self.quantiles = np.linspace(0, 1, n_quantiles + 1)
        self.values_quantiles = np.quantile(values, self.quantiles)

    def to_quantiles(self, values):
        quantile_scores = np.digitize(values, self.values_quantiles[1:-1])
        return quantile_scores

    def to_values(self, quantile_scores):
        values_from_scores = np.interp(
            quantile_scores, range(self.n_quantiles + 1), self.values_quantiles
        )
        return values_from_scores


class AskTellFewShotMulti:
    def __init__(
        self,
        prompt_template: PromptTemplate = None,
        suffix: Optional[str] = None,
        model: str = "text-curie-001",
        temperature: Optional[float] = None,
        prefix: Optional[str] = None,
        x_formatter: Callable[[str], str] = lambda x: x,
        y_formatter: Callable[[float], str] = lambda y: f"{y:0.2f}",
        y_name: str = "output",
        x_name: str = "input",
        selector_k: Optional[int] = None,
        k: int = 5,
        use_quantiles: bool = False,
        n_quantiles: int = 100,
        verbose: bool = False,
        cos_sim: bool = False,
    ) -> None:
        """Initialize Ask-Tell optimizer.

        You can pass formatters that will make your data more compatible with the model. Note that
        y as output form the model must be a float(can be parsed with ``float(y_str)``)

        Args:
            prompt_template: Prompt template that should take x and y (for few shot templates)
            suffix: Matching suffix for first part of prompt template - for actual completion.
            model: OpenAI base model to use for training and inference.
            temperature: Temperature to use for inference. If None, will use model default.
            prefix: Prefix to add before all examples (e.g., some context for the model).
            x_formatter: Function to format x for prompting.
            y_formatter: Function to format y for prompting.
            y_name: Name of y variable in prompt template (e.g., density, value of function, etc.)
            x_name: Name of x variable in prompt template (e.g., input, x, etc.). Only appears in inverse prompt
            selector_k: What k to use when switching to selection mode. If None, will use all examples
            k: Number of examples to use for each prediction.
            verbose: Whether to print out debug information.
        """
        self._selector_k = selector_k
        self._ready = False
        self._ys = []
        self.format_x = x_formatter
        self.format_y = y_formatter
        self._y_name = y_name
        self._x_name = x_name
        self._prompt_template = prompt_template
        self._suffix = suffix
        self._prefix = prefix
        self._model = model
        self._example_count = 0
        self._temperature = temperature
        self._k = k
        self._answer_choices = _answer_choices[:k]
        self.use_quantiles = use_quantiles
        self.n_quantiles = n_quantiles
        self._calibration_factor = None
        self._verbose = verbose
        self.tokens_used = 0
        self.cos_sim = cos_sim

    def _setup_llm(self, model: str, temperature: Optional[float] = None):
        return get_llm(
            model_name=model,
            # put stop with suffix, so it doesn't start babbling
            stop=[self.prompt.suffix.split()[0], self.inv_prompt.suffix.split()[0]],
            max_tokens=256,
            logprobs=self._k,
            temperature=0.05 if temperature is None else temperature,
        )

    def _setup_inv_llm(self, model: str, temperature: Optional[float] = None):
        return get_llm(
            model_name=model,
            # put stop with suffix, so it doesn't start babbling
            stop=[
                self.prompt.suffix.split()[0],
                self.inv_prompt.suffix.split()[0],
                # "\n",
            ],
            max_tokens=256,
            temperature=0.05 if temperature is None else temperature,
        )

    def _setup_inverse_prompt(self, example: Dict):
        prompt_template = PromptTemplate(
            input_variables=["x", "y", "y_name", "x_name"],
            template="If {y_name} is {y}, then {x_name} is {x}\n\n",
        )
        if example is not None:
            prompt_template.format(**example)
            examples = [example]
        else:
            examples = []
        example_selector = None
        if self._selector_k is not None:
            if len(examples) == 0:
                raise ValueError("Cannot do zero-shot with selector")
            if not self.cos_sim:
                example_selector = (
                    example_selector
                ) = MaxMarginalRelevanceExampleSelector.from_examples(
                    [example],
                    OpenAIEmbeddings(),
                    FAISS,
                    k=self._selector_k,
                )
            else:
                example_selector = (
                    example_selector
                ) = SemanticSimilarityExampleSelector.from_examples(
                    [example],
                    OpenAIEmbeddings(),
                    Chroma,
                    k=self._selector_k,
                )
        return FewShotPromptTemplate(
            examples=examples if example_selector is None else None,
            example_prompt=prompt_template,
            example_selector=example_selector,
            suffix="If {y_name} is {y}, then {x_name} is ",
            input_variables=["y", "y_name", "x_name"],
        )

    def _setup_prompt(
        self,
        example: Dict,
        prompt_template: Optional[PromptTemplate] = None,
        suffix: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> FewShotPromptTemplate:
        if prefix is None:
            prefix = f"For each multiple choice question, select correct choice from {','.join(self._answer_choices)}\n"
        if prompt_template is None:
            prompt_template = PromptTemplate(
                input_variables=["x", "Answer", "y_name"] + self._answer_choices,
                template="Q: Given {x}. What is {y_name}?\n"
                + "\n".join([f"{a}. {{{a}}}" for a in self._answer_choices])
                + "\nAnswer: {Answer}\n\n",
            )
            if suffix is not None:
                raise ValueError(
                    "Cannot provide suffix if using default prompt template."
                )
            suffix = "Q: Given {x}, what is {y_name}?\nA."
        elif suffix is None:
            raise ValueError("Must provide suffix if using custom prompt template.")
        # test out prompt
        if example is not None:
            prompt_template.format(**example)
            examples = [example]
        # TODO: make fake example text
        else:
            examples = []
        example_selector = None
        if self._selector_k is not None:
            if len(examples) == 0:
                raise ValueError("Cannot do zero-shot with selector")
            if not self.cos_sim:
                example_selector = (
                    example_selector
                ) = MaxMarginalRelevanceExampleSelector.from_examples(
                    [example],
                    OpenAIEmbeddings(),
                    FAISS,
                    k=self._selector_k,
                )
            else:
                example_selector = (
                    example_selector
                ) = SemanticSimilarityExampleSelector.from_examples(
                    [example],
                    OpenAIEmbeddings(),
                    Chroma,
                    k=self._selector_k,
                )
        return FewShotPromptTemplate(
            examples=examples if example_selector is None else None,
            example_prompt=prompt_template,
            example_selector=example_selector,
            suffix=suffix,
            prefix=prefix,
            input_variables=["x", "y_name"],
        )

    def _tell(self, x: str, y: float, alt_ys: Optional[List[float]] = None) -> Dict:
        # implementation of tell
        if alt_ys is not None:
            if len(alt_ys) != len(self._answer_choices) - 1:
                raise ValueError("Must provide 4 alternative ys.")
            alt_ys = [self.format_y(alt_y) for alt_y in alt_ys]
        else:
            alt_ys = []
            alt_y = y
            for i in range(100):
                if len(alt_ys) == len(self._answer_choices) - 1:
                    break
                if i < 50:
                    alt_y = y * 10 ** np.random.normal(0, 0.2)
                else:  # try something different
                    alt_y = y + np.random.uniform(-10, 10)
                if self.format_y(alt_y) not in alt_ys and self.format_y(
                    alt_y
                ) != self.format_y(y):
                    alt_ys.append(self.format_y(alt_y))
        # choose answer
        answer = np.random.choice(self._answer_choices)
        example_dict = dict(
            x=self.format_x(x),
            Answer=answer,
            y_name=self._y_name,
        )
        for a in self._answer_choices:
            if a == answer:
                example_dict[a] = self.format_y(y)
            else:
                example_dict[a] = alt_ys.pop()
        self._ys.append(y)
        inv_example = dict(
            x=self.format_x(x),
            y_name=self._y_name,
            y=self.format_y(y),
            x_name=self._x_name,
        )
        return example_dict, inv_example

    def tell(self, x: str, y: float, alt_ys: Optional[List[float]] = None) -> None:
        """Tell the optimizer about a new example."""
        example_dict, inv_example = self._tell(x, y, alt_ys)
        # we want to have example
        # to initialize prompts, so send it
        if not self._ready:
            self.prompt = self._setup_prompt(
                example_dict, self._prompt_template, self._suffix, self._prefix
            )
            self.inv_prompt = self._setup_inverse_prompt(inv_example)
            self.llm = self._setup_llm(self._model, self._temperature)
            self.inv_llm = self._setup_inv_llm(self._model, self._temperature)
            self._ready = True
        else:
            # in else, so we don't add twice
            if self._selector_k is not None:
                self.prompt.example_selector.add_example(example_dict)
                self.inv_prompt.example_selector.add_example(inv_example)
            else:
                self.prompt.examples.append(example_dict)
                self.inv_prompt.examples.append(inv_example)
        self._example_count += 1

    def _predict(self, queries: List[str]) -> List[DiscreteDist]:
        return openai_choice_predict(queries, self.llm, self._verbose)

    def inv_predict(self, y: float) -> str:
        """A rough inverse model"""
        if not self._ready:
            raise ValueError(
                "Must tell at least one example before inverse predicting."
            )
        query = self.inv_prompt.format(
            y=self.format_y(y), y_name=self._y_name, x_name=self._x_name
        )
        query = wrap_chatllm(query, self.inv_llm)
        x = self.inv_llm(query)
        if type(x) != str:
            return x.content
        return x

    def set_calibration_factor(self, calibration_factor):
        self._calibration_factor = calibration_factor

    def predict(self, x: str) -> Union[Tuple[float, float], List[Tuple[float, float]]]:
        """Predict the probability distribution and values for a given x.

        Args:
            x: The x value(s) to predict.
        Returns:
            The probability distribution and values for the given x.

        """
        if not isinstance(x, list):
            x = [x]
        if not self._ready:
            # special zero-shot
            self.prompt = self._setup_prompt(
                None, self._prompt_template, self._suffix, self._prefix
            )
            self.inv_prompt = self._setup_inverse_prompt(None)
            self.llm = self._setup_llm(self._model)
            self._ready = True

        if self._selector_k is not None:
            # have to update this until my PR is merged
            self.prompt.example_selector.k = min(self._example_count, self._selector_k)

        queries = [
            self.prompt.format(
                x=self.format_x(x_i),
                y_name=self._y_name,
            )
            for x_i in x
        ]
        results, tokens = self._predict(queries)
        self.tokens_used += tokens

        # need to replace any GaussDist with pop std
        for i, result in enumerate(results):
            if len(self._ys) > 1:
                ystd = np.std(self._ys)
            elif len(self._ys) == 1:
                ystd = self._ys[0]
            else:
                ystd = 10
            if isinstance(result, GaussDist):
                results[i].set_std(ystd)

        if self._calibration_factor:
            for i, result in enumerate(results):
                if isinstance(result, GaussDist):
                    results[i].set_std(result.std() * self._calibration_factor)
                elif isinstance(result, DiscreteDist):
                    results[i] = GaussDist(
                        results[i].mean(),
                        results[i].std() * self._calibration_factor,
                    )

        # compute mean and standard deviation
        if len(x) == 1:
            return results[0]
        return results

    def ask(
        self,
        possible_x: Union[Pool, List[str]],
        aq_fxn: str = "upper_confidence_bound",
        k: int = 1,
        inv_filter: int = 16,
        _lambda: float = 0.5,
    ) -> Tuple[List[str], List[float], List[float]]:
        """Ask the optimizer for the next x to try.



        Args:
            possible_x: List of possible x values to choose from.
            aq_fxn: Acquisition function to use.
            k: Number of x values to return.
            inv_filter: Reduce pool size to this number with inverse model. If 0, not used
            _lambda: Lambda value to use for UCB
        Return:
            The selected x values, their acquisition function values, and the predicted y modes.
            Sorted by acquisition function value (descending)
        """
        if type(possible_x) == type([]):
            possible_x = Pool(possible_x, self.format_x)

        if aq_fxn == "probability_of_improvement":
            aq_fxn = probability_of_improvement
        elif aq_fxn == "expected_improvement":
            aq_fxn = expected_improvement
        elif aq_fxn == "upper_confidence_bound":
            aq_fxn = partial(upper_confidence_bound, _lambda=_lambda)
        elif aq_fxn == "greedy":
            aq_fxn = greedy
        elif aq_fxn == "random":
            return (
                possible_x.sample(k),
                [0] * k,
                [0] * k,
            )
        else:
            raise ValueError(f"Unknown acquisition function: {aq_fxn}")

        if len(self._ys) == 0:
            best = 0
        else:
            best = np.max(self._ys)
        if inv_filter != 0 and inv_filter < len(possible_x):
            # not sure if this is too important,
            # but to get some diversity we
            # take last added y AND best
            approx_x = self.inv_predict(best * np.random.normal(1.0, 0.05))
            possible_x_l = possible_x.approx_sample(approx_x, inv_filter)

        else:
            possible_x_l = list(possible_x)
        results = self._ask(possible_x_l, best, aq_fxn, k)
        if len(results[0]) == 0 and len(possible_x_l) != 0:
            # if we have nothing, just return random one
            return (
                possible_x.sample(k),
                [0] * k,
                [0] * k,
            )
        return results

    def _ask(
        self, possible_x: List[str], best: float, aq_fxn: Callable, k: int
    ) -> Tuple[List[str], List[float], List[float]]:
        results = self.predict(possible_x)
        # drop empties
        results = [r for r in results if len(r) > 0]
        aq_vals = [aq_fxn(r, best) for r in results]
        selected = np.argsort(aq_vals)[::-1][:k]
        means = [r.mean() for r in results]
        return (
            [possible_x[i] for i in selected],
            [aq_vals[i] for i in selected],
            [means[i] for i in selected],
        )


class AskTellFewShotTopk(AskTellFewShotMulti):
    def _predict(self, queries: List[str]) -> List[DiscreteDist]:
        result, token_usage = openai_topk_predict(queries, self.llm, self._verbose)
        if self.use_quantiles and self.qt is None:
            raise ValueError(
                "Can't use quantiles without building the quantile transformer"
            )
        if self.use_quantiles:
            for r in result:
                if isinstance(r, GaussDist):
                    r._mean = self.qt.to_values(r._mean)
                elif isinstance(r, DiscreteDist):
                    r.values = self.qt.to_values(r.values)
        return result, token_usage

    def _setup_llm(self, model: str, temperature: Optional[float] = None):
        # nucleus sampling seems to get more diversity
        return get_llm(
            n=self._k,
            best_of=self._k,
            temperature=0.1 if temperature is None else temperature,
            model_name=model,
            top_p=1.0,
            stop=["\n", "###", "#", "##"],
            logit_bias={
                "198": -100,  # new line,
                "628": -100,  # double new line,
                "50256": -100,  # endoftext
            },
            max_tokens=256,
            logprobs=1,
        )

    def _setup_prompt(
        self,
        example: Dict,
        prompt_template: Optional[PromptTemplate] = None,
        suffix: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> FewShotPromptTemplate:
        if prefix is None:
            prefix = (
                "The following are correctly answered questions. "
                "Each answer is numeric and ends with ###\n"
            )
        if prompt_template is None:
            prompt_template = PromptTemplate(
                input_variables=["x", "y", "y_name"],
                template="Q: Given {x}, what is {y_name}?\nA: {y}###\n\n",
            )
            if suffix is not None:
                raise ValueError(
                    "Cannot provide suffix if using default prompt template."
                )
            suffix = "Q: Given {x}. What is {y_name}?\nA: "
        elif suffix is None:
            raise ValueError("Must provide suffix if using custom prompt template.")
        # test out prompt
        if example is not None:
            prompt_template.format(**example)
            examples = [example]
        # TODO: make fake example text
        else:
            examples = []
        example_selector = None
        if self._selector_k is not None:
            if len(examples) == 0:
                raise ValueError("Cannot do zero-shot with selector")
            if not self.cos_sim:
                example_selector = (
                    example_selector
                ) = MaxMarginalRelevanceExampleSelector.from_examples(
                    [example],
                    OpenAIEmbeddings(),
                    FAISS,
                    k=self._selector_k,
                )
            else:
                example_selector = (
                    example_selector
                ) = SemanticSimilarityExampleSelector.from_examples(
                    [example],
                    OpenAIEmbeddings(),
                    Chroma,
                    k=self._selector_k,
                )
        return FewShotPromptTemplate(
            examples=examples if example_selector is None else None,
            example_prompt=prompt_template,
            example_selector=example_selector,
            suffix=suffix,
            prefix=prefix,
            input_variables=["x", "y_name"],
        )

    def _tell(self, x: str, y: float, alt_ys: Optional[List[float]] = None) -> Dict:
        """Tell the optimizer about a new example."""

        if self.use_quantiles:
            self.qt = QuantileTransformer(
                values=self._ys + [y], n_quantiles=self.n_quantiles
            )
            y = self.qt.to_quantiles(y)

        if alt_ys is not None:
            raise ValueError("Alt ys not supported for topk.")
        example_dict = dict(
            x=self.format_x(x),
            y=self.format_y(y),
            y_name=self._y_name,
        )
        self._ys.append(y)
        inv_dict = dict(
            x=self.format_x(x),
            y=self.format_y(y),
            y_name=self._y_name,
            x_name=self._x_name,
        )
        return example_dict, inv_dict



import numpy as np

from typing import *
from .asktell import AskTellFewShotTopk
from .llm_model import GaussDist
from langchain.prompts.few_shot import FewShotPromptTemplate
from langchain.prompts.prompt import PromptTemplate
from langchain.vectorstores import Chroma
from langchain.embeddings import OpenAIEmbeddings
from langchain.prompts.example_selector import SemanticSimilarityExampleSelector


class AskTellNearestNeighbor(AskTellFewShotTopk):
    def __init__(self, knn=1, **kwargs):
        super().__init__(selector_k=knn, **kwargs)
        self.knn = knn

    def _tell(self, x: str, y: float, alt_ys: Optional[List[float]] = None) -> Dict:
        """Tell the optimizer about a new example."""
        if alt_ys is not None:
            raise ValueError("Alt ys not supported for topk.")
        example_dict = dict(
            x=self.format_x(x),
            y=self.format_y(y),
            y_name=self._y_name,
        )
        self._ys.append(y)
        return example_dict

    def tell(self, x: str, y: float, alt_ys: Optional[List[float]] = None) -> None:
        """Tell the optimizer about a new example."""
        example_dict = self._tell(x, y, alt_ys)
        # we want to have example
        # to initialize prompts, so send it
        if not self._ready:
            self.prompt = self._setup_prompt(
                example_dict, self._prompt_template, self._suffix, self._prefix
            )
            self._ready = True
        else:
            if self._selector_k is not None:
                self.prompt.example_selector.add_example(example_dict)
            else:
                self.prompt.examples.append(example_dict)
        self._example_count += 1

    def _setup_prompt(
        self,
        example: Dict,
        prompt_template: Optional[PromptTemplate] = None,
        suffix: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> FewShotPromptTemplate:
        if prefix is None:
            prefix = (
                "The following are correctly answered questions. "
                "Each answer is numeric and ends with ###\n"
            )
        if prompt_template is None:
            prompt_template = PromptTemplate(
                input_variables=["x", "y", "y_name"],
                template="Q: Given {x}, what is {y_name}?\nA: {y}###\n\n",
            )
            if suffix is not None:
                raise ValueError(
                    "Cannot provide suffix if using default prompt template."
                )
            suffix = "Q: Given {x}. What is {y_name}?\nA: "
        elif suffix is None:
            raise ValueError("Must provide suffix if using custom prompt template.")
        if example is not None:
            prompt_template.format(**example)
            examples = [example]
        else:
            examples = []
        example_selector = None
        if self._selector_k is not None:
            if len(examples) == 0:
                raise ValueError("Cannot do zero-shot with selector")
            example_selector = (
                example_selector
            ) = SemanticSimilarityExampleSelector.from_examples(
                [example],
                OpenAIEmbeddings(),
                Chroma,
                k=self._selector_k,
            )
        return FewShotPromptTemplate(
            examples=examples if example_selector is None else None,
            example_prompt=prompt_template,
            example_selector=example_selector,
            suffix=suffix,
            prefix=prefix,
            input_variables=["x", "y_name"],
        )

    def predict(self, x: str) -> Union[Tuple[float, float], List[Tuple[float, float]]]:
        if not isinstance(x, list):
            x = [x]
        if not self._ready:
            self.prompt = self._setup_prompt(
                None, self._prompt_template, self._suffix, self._prefix
            )
            self._ready = True

        if self._selector_k is not None:
            self.prompt.example_selector.k = min(self._example_count, self._selector_k)

        selected = [
            self.prompt.example_selector.select_examples({"x": x_i}) for x_i in x
        ]

        predictions = [[float(s["y"]) for s in selected_i] for selected_i in selected]

        results = [GaussDist(np.mean(p), np.std(p)) for p in predictions]

        if len(x) == 1:
            return results[0]
        return results



from bolift import BOLiftTool, Pool
import os


def test_tell_tools():
    pool_list = ["red", "green", "blue"]
    pool = Pool(pool_list)
    tool = BOLiftTool(pool)
    with open('test.csv', 'w') as f:
        f.write("yellow, 1\n")
        f.write("teal, 2\n")
    tool("Ask")
    s = tool("Tell test.csv")
    assert s == "Added 2 training examples."
    tool("Ask")
    os.remove("test.csv")


from bolift import llm_model
import bolift
import numpy as np

np.random.seed(0)


def test_completion():
    llm = llm_model.get_llm(stop=["\n\n"])
    assert llm("The value of 1 + 1 is") == " 2"


def test_parse_response():
    prompt = """
Problem 1: What is 4 x 5?
A. 12
B. 32
C. 20
D. 16
E. 24

Answer: C

Problem 2: What is 4 x 4?
"""
    llm = llm_model.get_llm(logprobs=5)
    generation = llm.generate([prompt]).generations[0][0]
    result = llm_model.parse_response(generation, prompt, llm)
    # make sure answer is max
    assert 16 in result.values.astype(int)


def test_parse_response_topk():
    prompt = "2 + 2 is"
    llm = llm_model.get_llm(
        n=3,
        best_of=3,
        temperature=1,
        model_name="text-babbage-001",
        top_p=0.99,
        stop=["\n"],
        logprobs=1,
    )
    g = llm.generate([prompt]).generations
    result = llm_model.parse_response_topk(g[0])
    # make sure answer is max
    assert abs(result.mode().astype(int) - 4) <= 1


def test_tell_fewshot():
    asktell = bolift.AskTellFewShotMulti(
        x_formatter=lambda x: f"y = 2 * {x}",
        y_formatter=lambda y: str(int(y)),
        verbose=True,
    )
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(3, 6)
    dist = asktell.predict(3)
    dist.mode()
    # assert m == 6


def test_tell_inv_fewshot():
    asktell = bolift.AskTellFewShotMulti(
        x_formatter=lambda x: f"2 * {x}",
        y_formatter=lambda y: str(int(y)),
        verbose=True,
    )
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(3, 6)
    inverse = asktell.inv_predict(6)
    assert "*" in inverse


def test_tell_inv_fewshot_topk():
    asktell = bolift.AskTellFewShotTopk(
        x_formatter=lambda x: f"2 * {x}",
        y_formatter=lambda y: str(int(y)),
        verbose=True,
    )
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(3, 6)
    inverse = asktell.inv_predict(6)
    assert "*" in inverse


def test_tell_fewshot_topk():
    asktell = bolift.AskTellFewShotTopk(x_formatter=lambda x: f"y = 2*{x}")
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(16, 32)
    dist = asktell.predict(2)
    m = dist.mode().astype(int)
    assert m == 4


def test_tell_fewshot_vark_topk():
    asktell = bolift.AskTellFewShotTopk(
        x_formatter=lambda x: f"y = 2 * {x}", k=1, y_formatter=lambda y: str(int(y))
    )
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    dist = asktell.predict(2)
    m = dist.mode()
    assert m == 4


def test_tell_fewshot_vark():
    asktell = bolift.AskTellFewShotMulti(
        x_formatter=lambda x: f"y = 2*{x}",
        k=3,
        verbose=True,
        y_formatter=lambda y: str(int(y)),
    )
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(3, 6)
    dist = asktell.predict(2)
    dist.mode()
    # assert m == 4


def test_tell_fewshot_selector():
    asktell = bolift.AskTellFewShotMulti(
        x_formatter=lambda x: f"y = 2 + {x}",
        selector_k=3,
        y_formatter=lambda y: str(int(y)),
        verbose=True,
    )
    for i in range(5):
        asktell.tell(i, 2 + i)
    dist = asktell.predict(3)
    assert abs(3 + 2 - dist.mode().astype(int)) < 10


def test_tell_fewshot_selector_less():
    asktell = bolift.AskTellFewShotMulti(
        x_formatter=lambda x: f"y = 2 + {x}",
        selector_k=10,
        y_formatter=lambda y: str(int(y)),
        verbose=True,
    )
    for i in range(5):
        asktell.tell(i, 2 + i)
    dist = asktell.predict(3)
    assert abs(3 + 2 - dist.mode().astype(int)) < 10


def test_tell_fewshot_topk_selector():
    asktell = bolift.AskTellFewShotTopk(
        x_formatter=lambda x: f"y = 2 + {x}",
        selector_k=3,
        y_formatter=lambda y: str(int(y)),
        verbose=True,
    )
    for i in range(5):
        asktell.tell(i, 2 + i)
    dist = asktell.predict(3)
    assert abs(3 + 2 - dist.mode().astype(int)) < 10


def test_ask_ei_fewshot():
    asktell = bolift.AskTellFewShotMulti(x_formatter=lambda x: f"y = 2 * {x}")
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(16, 32)
    best, _, _ = asktell.ask([2, 8])
    assert best[0] == 8


def test_ask_ei_fewshot_pool():
    asktell = bolift.AskTellFewShotMulti(x_formatter=lambda x: f"y = 2 * {x}")
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(16, 32)
    pool = list(range(30))
    best, _, _ = asktell.ask(pool, inv_filter=3)
    assert best[0] > 10


def test_ask_ei_fewshot_pool_topk():
    asktell = bolift.AskTellFewShotTopk(x_formatter=lambda x: f"y = 2 * {x}")
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(16, 32)
    pool = list(range(30))
    best, _, _ = asktell.ask(pool, inv_filter=3)
    assert best[0] > 10


def test_ask_ei_fewshot_topk():
    asktell = bolift.AskTellFewShotTopk(x_formatter=lambda x: f"y = 2 * {x}")
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(16, 32)
    best, _, _ = asktell.ask([2, 8])
    assert best[0] == 8


def test_ask_zero_fewshot():
    asktell = bolift.AskTellFewShotMulti(x_formatter=lambda x: f"y = 2 * {x}")
    _, scores, _ = asktell.ask([2, 8], k=2)
    assert scores[0] > scores[1]
    asktell.ask([2, 8], k=2, aq_fxn="greedy")


def test_ask_zero_fewshot_topk():
    asktell = bolift.AskTellFewShotTopk(
        x_formatter=lambda x: f"y = f({x}) for the shifted Ackley function",
        model="text-babbage-001",
    )
    _, scores, means = asktell.ask([2, 8], k=2)
    assert scores[0] >= scores[1]
    asktell.ask([2, 8], k=2, aq_fxn="greedy")


def test_ask_ei_fewshot_finetune():
    asktell = bolift.AskTellFinetuning(
        x_formatter=lambda x: f"y = 2 * {x}",
        finetune=False,
    )
    asktell.tell(2, 4)
    asktell.tell(1, 2)
    asktell.tell(16, 32)
    best, _, _ = asktell.ask([2, 8])
    assert best[0] == 8


def test_prepare_data_finetuning():
    import os

    asktell = bolift.AskTellFinetuning(
        x_formatter=lambda x: f"y = 2 * {x}", model="text-ada-001", finetune=True
    )
    prompts = ["2", "4", "8"]
    completions = ["4", "8", "16"]
    asktell.prepare_data(prompts, completions, "./test.jsonl")
    assert os.path.exists("./test.jsonl")
    assert open("./test.jsonl").readlines() == [
        '{"prompt": "2", "completion": "4"}\n',
        '{"prompt": "4", "completion": "8"}\n',
        '{"prompt": "8", "completion": "16"}\n',
    ]
    os.remove("./test.jsonl")


def test_upload_data_finetuning():
    import os
    import openai
    import time

    asktell = bolift.AskTellFinetuning(
        x_formatter=lambda x: f"y = 2 * {x}", model="text-ada-001", finetune=True
    )
    prompts = ["2", "4", "8"]
    completions = ["4", "8", "16"]
    asktell.prepare_data(prompts, completions, "./test.jsonl")
    file_id = asktell.upload_data("./test.jsonl")
    time.sleep(30)  # Sometimes it take a few seconds for the file to be uploaded
    assert file_id is not None
    assert (
        openai.File.retrieve(file_id).status == "uploaded"
        or openai.File.retrieve(file_id).status == "processed"
    )
    os.remove("./test.jsonl")
    openai.File.delete(file_id)


def test_llm_usage():
    asktell = bolift.AskTellFewShotMulti(
        x_formatter=lambda x: f"y = 2 + {x}",
        selector_k=10,
        y_formatter=lambda y: str(int(y)),
    )
    for i in range(5):
        asktell.tell(i, 2 + i)
    asktell.predict(3)
    print(asktell.tokens_used)
    assert asktell.tokens_used > 0


def test_gpr():
    asktell = bolift.AskTellGPR(
        x_formatter=lambda x: f"y = 2 + {x}",
        y_formatter=lambda y: str(int(y)),
    )
    for i in range(5):
        asktell.tell(i, 2 + i, train=False)
    asktell.tell(5, 7, train=True)
    asktell.predict(5000)
    assert asktell.ask([1, 5])[0][0] == 5



import numpy as np
import pandas as pd
from .pool import Pool
from .asktell import AskTellFewShotTopk
from .llm_model import GaussDist

from typing import *
from botorch.models.gp_regression import SingleTaskGP
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood
from botorch.optim.fit import fit_gpytorch_torch
import torch
from langchain.embeddings import OpenAIEmbeddings
from sklearn.manifold import Isomap


class AskTellGPR(AskTellFewShotTopk):
    def __init__(self, n_components=2, pool=None, cache_path=None, **kwargs):
        super().__init__(**kwargs)
        self._set_regressor()
        self.examples = []
        self._embedding = OpenAIEmbeddings()
        self._embeddings_cache = self._get_cache(cache_path)
        self.isomap = Isomap(n_components=n_components)
        self.pool = pool
        if self.pool is not None:
            self._initialize_isomap()

    def _initialize_isomap(self):
        pool_embeddings = self._query_cache(self.pool._available)
        self.isomap.fit(pool_embeddings)

    def _get_cache(self, cache_path=None):
        try:
            cache = pd.read_csv(cache_path)
            print(f"Loaded cache from {cache_path}.")
        except:
            print("Cached embeddings not found. Creating new cache table.")
            cache = pd.DataFrame({"x": [], "embedding": []})
        return cache

    def save_cache(self, cache_path):
        self._embeddings_cache.to_csv(cache_path, index=False)

    def _query_cache(self, X):
        in_cache = self._embeddings_cache["x"].to_list()
        not_in_cache = np.setdiff1d(X, in_cache)
        new_embeddings = []
        if not_in_cache.size > 0:
            new_embeddings = self._embedding.embed_documents(not_in_cache.tolist(), 250)
        self._embeddings_cache = pd.concat(
            [
                self._embeddings_cache,
                pd.DataFrame({"x": not_in_cache, "embedding": new_embeddings}),
            ],
            ignore_index=True,
        )
        embedding = [
            self._embeddings_cache[self._embeddings_cache["x"] == xi][
                "embedding"
            ].to_list()[0]
            for xi in X
        ]

        if len(embedding) != len(X):
            raise ValueError(
                (
                    "Embedding length does not match X length."
                    "Something went wrong on caching."
                )
            )

        return embedding

    def _set_regressor(self):
        self.likelihood = GaussianLikelihood()
        self.regressor = None

    def _predict(self, X):
        if len(X) == 0:
            raise ValueError("X is empty")
        embedding = self._query_cache(X)
        embedding_isomap = self.isomap.transform(embedding)
        results = []
        with torch.no_grad():
            self.regressor.eval()
            self.likelihood.eval()
            means = self.likelihood(self.regressor(torch.tensor(embedding_isomap))).mean
            stds = self.likelihood(
                self.regressor(torch.tensor(embedding_isomap))
            ).variance.sqrt()
        results = [GaussDist(mean.item(), std.item()) for mean, std in zip(means, stds)]
        return results, 0

    def _train(self, X, y):
        embedding = self._query_cache(X)
        embedding = np.array(embedding)
        if self.pool is None:
            embedding_isomap = self.isomap.fit_transform(embedding)
        else:
            embedding_isomap = self.isomap.transform(embedding)
        train_x = torch.tensor(embedding_isomap)
        train_y = torch.tensor(list(map(float, y))).unsqueeze(-1).double()
        self.regressor = SingleTaskGP(train_x, train_y)
        mll = ExactMarginalLogLikelihood(self.regressor.likelihood, self.regressor)
        fit_gpytorch_torch(mll)

    def ask(
        self,
        possible_x: Union[Pool, List[str]],
        aq_fxn: str = "upper_confidence_bound",
        k: int = 1,
        inv_filter: int = 16,
        _lambda: float = 0.5,
    ) -> Tuple[List[str], List[float], List[float]]:
        # just have this here to override default
        return super().ask(possible_x, aq_fxn, k, 0, _lambda)

    def tell(
        self, x: str, y: float, alt_ys: Optional[List[float]] = None, train=True
    ) -> None:
        """Tell the optimizer about a new example."""
        example_dict, inv_example = self._tell(x, y, alt_ys)
        # we want to have example
        # to initialize prompts, so send it
        if not self._ready:
            self.prompt = self._setup_prompt(
                None, self._prompt_template, self._suffix, self._prefix
            )
            self.inv_prompt = self._setup_inverse_prompt(inv_example)
            self.llm = self._setup_llm(self._model, self._temperature)
            self.inv_llm = self._setup_inv_llm(self._model, self._temperature)
            self._ready = True

        self.examples.append(example_dict)

        if train:
            self._train(
                [
                    self.prompt.format(
                        x=ex["x"],
                        y_name=self._y_name,
                    )
                    for ex in self.examples
                ],
                [ex["y"] for ex in self.examples],
            )

    def predict(self, x: str) -> Union[Tuple[float, float], List[Tuple[float, float]]]:
        """Predict the probability distribution and values for a given x.

        Args:
            x: The x value(s) to predict.
        Returns:
            The probability distribution and values for the given x.

        """
        if not isinstance(x, list):
            x = [x]
        if not self._ready:
            # special zero-shot
            self.prompt = self._setup_prompt(
                None, self._prompt_template, self._suffix, self._prefix
            )
            self.inv_prompt = self._setup_inverse_prompt(None)
            self.llm = self._setup_llm(self._model)
            self._ready = True

        if self._selector_k is not None:
            # have to update this until my PR is merged
            self.prompt.example_selector.k = min(self._example_count, self._selector_k)

        queries = [
            self.prompt.format(
                x=self.format_x(x_i),
                y_name=self._y_name,
            )
            for x_i in x
        ]
        results, tokens = self._predict(queries)
        self.tokens_used += tokens

        # compute mean and standard deviation
        if len(x) == 1:
            return results[0]
        return results



import numpy as np
import re
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from langchain.callbacks import get_openai_callback
from langchain.cache import InMemoryCache
import langchain
from dataclasses import dataclass

from langchain.schema import HumanMessage, SystemMessage


def wrap_chatllm(query_list, llm):
    if type(llm) == ChatOpenAI:
        system_message_prompt = SystemMessage(
            content="You are a bot that can predict chemical and material properties. Do not explain answers, just provide numerical predictions."
        )
        if type(query_list) == str:
            query_list = [system_message_prompt, HumanMessage(content=query_list)]
        else:
            query_list = [
                [system_message_prompt, HumanMessage(content=q)] for q in query_list
            ]
    return query_list


@dataclass
class DiscreteDist:
    values: np.ndarray
    probs: np.ndarray

    def __post_init__(self):
        # make sure np arrays
        self.values = np.array(self.values)
        self.probs = np.array(self.probs)
        uniq_values = np.unique(self.values)
        if len(uniq_values) < len(self.values):
            # need to mergefg
            uniq_probs = np.zeros(len(uniq_values))
            for i, v in enumerate(uniq_values):
                uniq_probs[i] = np.sum(self.probs[self.values == v])
            self.values = uniq_values
            self.probs = uniq_probs

    def sample(self):
        return np.random.choice(self.values, p=self.probs)

    def mean(self):
        return np.sum(self.values * self.probs)

    def mode(self):
        return self.values[np.argmax(self.probs)]

    def std(self):
        return np.sqrt(np.sum((self.values - self.mean()) ** 2 * self.probs))

    def __repr__(self):
        return f"DiscreteDist({self.values}, {self.probs})"

    def __len__(self):
        return len(self.values)


@dataclass
class GaussDist:
    _mean: float
    _std: float

    def sample(self):
        return np.random.normal(self._mean, self._std)

    def mean(self):
        return self._mean

    def mode(self):
        return self._mean

    def std(self):
        return self._std

    def set_std(self, value):
        self._std = value

    def __repr__(self):
        return f"GaussDist({self._mean}, {self._std})"

    def __len__(self):
        return 1


def make_dd(values, probs):
    dd = DiscreteDist(values, probs)
    if len(dd) == 1:
        return GaussDist(dd.mean(), None)
    return dd


langchain.llm_cache = InMemoryCache()


def get_llm(
    model_name="text-babbage-001",
    temperature=0.05,
    n=1,
    top_p=1,
    best_of=1,
    max_tokens=128,
    **kwargs,
):
    if model_name in ["gpt-4", "gpt-3.5-turbo"]:
        if "logprobs" in kwargs:
            # not supported
            del kwargs["logprobs"]
        return ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            n=n,
            model_kwargs=kwargs,
            max_tokens=max_tokens,
        )

    return OpenAI(
        model_name=model_name,
        temperature=temperature,
        n=n,
        best_of=best_of,
        top_p=top_p,
        model_kwargs=kwargs,
        max_tokens=max_tokens,
    )


def parse_response(generation, prompt, llm):
    # first parse the options into numbers
    text = generation.text
    matches = re.findall(r"([A-Z])\. .*?([\+\-\d][\d\.e]*)", text)
    values = dict()
    k = None
    for m in matches:
        try:
            k, v = m[0], float(m[1])
            values[k] = v
        except ValueError:
            pass
        k = None
    # now get log prob of tokens after Answer:
    tokens = generation.generation_info["logprobs"]["top_logprobs"]
    offsets = generation.generation_info["logprobs"]["text_offset"]
    if "Answer:" not in text:
        # try to extend
        c_generation = llm.generate([prompt + text + "\nAnswer:"]).generations[0][0]

        logprobs = c_generation.generation_info["logprobs"]["top_logprobs"][0]
    else:
        # find token probs for answer
        # feel like this is supper brittle, but not sure what else to try
        at_answer = False
        for i in range(len(offsets)):
            start = offsets[i] - offsets[0]
            end = offsets[i + 1] - offsets[0] if i < len(offsets) - 1 else -1
            selected_token = text[start:end]
            if "Answer" in selected_token:
                at_answer = True
            if at_answer and selected_token.strip() in values:
                break
        logprobs = tokens[i]
    result = [
        (values[k.strip()], v) for k, v in logprobs.items() if k.strip() in values
    ]
    probs = np.exp(np.array([v for k, v in result]))
    probs = probs / np.sum(probs)
    # return DiscreteDist(np.array([k for k, v in result]), probs)
    return make_dd(np.array([k for k, v in result]), probs)


def truncate(s):
    """Truncate to first number"""
    try:
        return re.findall(r"[-+]?\d*\.\d+|\d+", s)[0]
    except IndexError:
        return s


def remove_overlap(s1, s2, check_l=10):
    """There may be some of s1 in s2. Remove it and return rest of s2"""
    for i in range(check_l, 0, -1):
        if s1[-i:] == s2[:i]:
            return s2[len(s1[-i:]) :]
    return s2


def parse_response_topk(generations):
    values, logprobs = [], []
    for gen in generations:
        try:
            v = float(truncate(gen.text))
            values.append(v)
        except ValueError:
            continue
        # can do inner sum because there is only one token
        lp = sum(
            [
                sum(x.to_dict().values())
                for x in gen.generation_info["logprobs"]["top_logprobs"]
            ]
        )
        logprobs.append(lp)

    probs = np.exp(np.array(logprobs))
    probs = probs / np.sum(probs)
    # return DiscreteDist(np.array(values), probs)
    return make_dd(np.array(values), probs)


def parse_response_n(generations):
    values = []
    for gen in generations:
        try:
            v = float(truncate(gen.text))
            values.append(v)
        except ValueError:
            continue

    probs = [1 / len(values) for _ in values]
    # return DiscreteDist(np.array(values), probs)
    return make_dd(np.array(values), probs)


def openai_choice_predict(query_list, llm, verbose, *args, **kwargs):
    """Predict the output numbers for a given list of queries"""
    with get_openai_callback() as cb:
        completion_response = llm.generate(query_list, *args, **kwargs)
        token_usage = cb.total_tokens
    if verbose:
        print("-" * 80)
        print(query_list[0])
        print("-" * 80)
        print(query_list[0] + completion_response.generations[0][0].text)
        print("-" * 80)
    results = []
    for gen, q in zip(completion_response.generations, query_list):
        results.append(parse_response(gen[0], q, llm))
    return results, token_usage


def openai_topk_predict(query_list, llm, verbose, *args, **kwargs):
    """Predict the output numbers for a given list of queries"""
    query_list = wrap_chatllm(query_list, llm)
    with get_openai_callback() as cb:
        completion_response = llm.generate(query_list, *args, **kwargs)
        token_usage = cb.total_tokens
    if verbose:
        print("-" * 80)
        print(query_list[0])
        print("-" * 80)
        print(query_list[0] + completion_response.generations[0][0].text)
        print("-" * 80)
    results = []
    for gens in completion_response.generations:
        if type(llm) == ChatOpenAI:
            results.append(parse_response_n(gens))
        else:
            results.append(parse_response_topk(gens))
    return results, token_usage



import numpy as np
from scipy.stats import norm
from .llm_model import DiscreteDist, GaussDist


def expected_improvement(dist, best):
    """Expected improvement for the given discrete distribution"""
    if isinstance(dist, DiscreteDist):
        return expected_improvement_d(dist.probs, dist.values, best)
    elif isinstance(dist, GaussDist):
        return expected_improvement_g(dist.mean(), dist.std(), best)


def probability_of_improvement(dist, best):
    """Probability of improvement for the given discrete distribution"""
    if isinstance(dist, DiscreteDist):
        return probability_of_improvement_d(dist.probs, dist.values, best)
    elif isinstance(dist, GaussDist):
        return probability_of_improvement_g(dist.mean(), dist.std(), best)


def upper_confidence_bound(dist, best, _lambda):
    """Upper confidence bound for the given discrete distribution"""
    if isinstance(dist, DiscreteDist):
        return upper_confidence_bound_d(dist.probs, dist.values, best, _lambda)
    elif isinstance(dist, GaussDist):
        return upper_confidence_bound_g(dist.mean(), dist.std(), best, _lambda)


def greedy(dist, best):
    """Greedy selection (most likely point) for the given discrete distribution"""
    if isinstance(dist, DiscreteDist):
        return greedy_d(dist.probs, dist.values, best)
    elif isinstance(dist, GaussDist):
        return greedy_g(dist.mean(), dist.std(), best)


def expected_improvement_d(probs, values, best):
    """Expected improvement for the given discrete distribution"""
    ei = np.sum(np.maximum(values - best, 0) * probs)
    return ei


def probability_of_improvement_d(probs, values, best):
    """Probability of improvement for the given discrete distribution"""
    pi = np.sum(np.cast[float](values > best) * probs)
    return pi


def upper_confidence_bound_d(probs, values, best, _lambda):
    """Upper confidence bound for the given discrete distribution"""
    mu = np.sum(values * probs)
    sigma = np.sqrt(np.sum((values - mu) ** 2 * probs))
    return mu + _lambda * sigma


def greedy_d(probs, values, best):
    """Greedy selection (most likely point) for the given discrete distribution"""
    return values[np.argmax(probs)]


def expected_improvement_g(mean, std, best):
    """Expected improvement for the given Gaussian distribution"""
    z = (mean - best) / std
    ei = (mean - best) * norm.cdf(z) + std * norm.pdf(z)
    return ei


def probability_of_improvement_g(mean, std, best):
    """Probability of improvement for the given Gaussian distribution"""
    z = (mean - best) / std
    pi = norm.cdf(z)
    return pi


def upper_confidence_bound_g(mean, std, best, _lambda):
    """Upper confidence bound for the given Gaussian distribution"""
    return mean + _lambda * std


def greedy_g(mean, std, best):
    """Greedy selection (most likely point) for the given Gaussian distribution"""
    return mean



import os
import openai
import pandas as pd
import time
import json

from typing import *
from .asktell import AskTellFewShotTopk
from langchain.prompts.prompt import PromptTemplate


class AskTellFinetuning(AskTellFewShotTopk):
    def __init__(
        self,
        prompt_template: PromptTemplate = None,
        suffix: Optional[str] = None,
        model: str = "text-curie-001",
        temperature: Optional[float] = None,
        prefix: Optional[str] = None,
        x_formatter: Callable[[str], str] = lambda x: x,
        y_formatter: Callable[[float], str] = lambda y: f"{y:0.2f}",
        y_name: str = "y",
        selector_k: Optional[int] = None,
        k: int = 5,
        verbose: bool = False,
        id: str = None,
        n_epochs: int = 8,
        learning_rate_multiplier: int = 0.02,
        finetune: bool = False,
        examples: List[Tuple[str, float]] = [],
    ) -> None:
        super().__init__(
            prompt_template=prompt_template,
            suffix=suffix,
            model=model,
            temperature=temperature,
            prefix=prefix,
            x_formatter=x_formatter,
            y_formatter=y_formatter,
            y_name=y_name,
            selector_k=selector_k,
            k=k,
            verbose=verbose,
        )
        self.n_epochs = n_epochs
        self.learning_rate_multiplier = learning_rate_multiplier
        self.response = None
        self.finetune = finetune
        self.base_model = model.split("-")[1]
        self.examples = examples
        if id:
            self.response = openai.FineTune.retrieve(id=id)
            self._model = self.response["fine_tuned_model"]
            self.base_model = self.response["fine_tuned_model"]
        # self.llm = self._setup_llm(model=self.base_model)

    def prepare_data_from_file(self, inFile, outFile):
        reg_df = pd.read_csv(inFile)
        with open(outFile, "w") as f:
            for e in range(len(reg_df)):
                f.write(
                    f'{{"prompt":"{reg_df.iloc[e]["prompt"]}", "completion":"{reg_df.iloc[e]["completion"]}"}}\n'
                )

    def prepare_data(self, prompts, completions, outFile):
        with open(outFile, "w") as f:
            for p, c in zip(prompts, completions):
                p = p.replace("°", "")
                f.write(f'{{"prompt": "{p}", "completion": "{c}"}}\n')

    def upload_data(self, data):
        f = openai.File.create(
            file=open(data, "r"),
            purpose="fine-tune",
            user_provided_filename="FTimplementation.jsonl",
        )
        file_id = f["id"]
        return file_id

    def create_fine_tune(self, file_id, base_model):
        response = openai.FineTune.create(
            training_file=file_id,
            model=base_model,
            n_epochs=self.n_epochs,
            learning_rate_multiplier=self.learning_rate_multiplier,
        )
        return response

    def fine_tune(self, prompts, completions, out_path="./out", out_file=None) -> None:
        if not os.path.exists(f"{out_path}"):
            os.makedirs(f"{out_path}")
        self.prepare_data(
            prompts,
            completions,
            f"{out_path}/train_data_{len(self.prompt.examples)}.jsonl",
        )
        file_id = self.upload_data(
            f"{out_path}/train_data_{len(self.prompt.examples)}.jsonl"
        )

        response = self.create_fine_tune(file_id, self.base_model)

        s = openai.FineTune.retrieve(id=response["id"]).status
        t = 0
        while s != "succeeded":
            if t % 3 == 0:
                s += ".   "
            elif t % 3 == 1:
                s += "..  "
            elif t % 3 == 2:
                s += "... "
            event_message = openai.FineTune.retrieve(id=response["id"]).events[-1][
                "message"
            ]
            s += f"{event_message}                                     "
            print(s, end="\r")

            s = openai.FineTune.retrieve(id=response["id"]).status
            t += 1
            time.sleep(2)
        print("\n")

        self.response = openai.FineTune.retrieve(id=response["id"])
        self.base_model = self.response["fine_tuned_model"]
        self.llm = self._setup_llm(model=self.response["fine_tuned_model"])
        if out_file:
            with open(f"{out_path}/{out_file}.dat", "w") as out:
                out.write(json.dumps(self.response, indent=4))
        else:
            with open(
                f"{out_path}/FT_model_{len(self.prompt.examples)}.dat", "w"
            ) as out:
                out.write(json.dumps(self.response, indent=4))

    def get_model_name(self):
        return self.response["fine_tuned_model"] if self.response else self.base_model

    def _tell(self, x: str, y: float, alt_ys: Optional[List[float]] = None) -> Dict:
        """Tell the optimizer about a new example."""
        if alt_ys is not None:
            raise ValueError("Alt ys not supported for topk.")

        if self.finetune:
            if len(self.prompt.examples) % 5 == 0 and len(self.prompt.examples) > 0:
                self.examples.extend(self.prompt.examples)
                prompts = [
                    f"Q: Given {p['x']}. What is {self._y_name}?\\nA: "
                    for p in self.examples
                ]
                completions = [p["y"] for p in self.examples]
                self.fine_tune(prompts, completions)
                self.prompt.examples = None
        example_dict = dict(
            x=self.format_x(x),
            y=self.format_y(y),
            y_name=self._y_name,
        )
        self._ys.append(y)
        inv_dict = dict(
            x=self.format_x(x),
            y=self.format_y(y),
            y_name=self._y_name,
            x_name=self._x_name,
        )
        return example_dict, inv_dict



"""utilities for building and selecting from a pool"""
from typing import List, Any, Callable
import numpy as np
from langchain.vectorstores import FAISS
from langchain.embeddings import OpenAIEmbeddings


class Pool:
    """Class for sampling from pool of possible data points

    Example:
        >>> pool = Pool(['a', 'b', 'c', 'd', 'e'])
        >>> pool.sample(3)
        ['a', 'd', 'c']
        >>> pool.choose('a')
        >>> pool.sample(3)
        ['b', 'c', 'd']
        >>> pool.approx_sample('a', 3)
        ['b', 'c', 'd']
    """

    def __init__(self, pool: List[Any], formatter: Callable = lambda x: str(x)) -> None:
        if type(pool) is not list:
            raise TypeError("Pool must be a list")
        self._pool = pool
        self._selected = []
        self._available = pool[:]
        self.format = formatter
        self._db = FAISS.from_texts(
            [formatter(x) for x in pool],
            OpenAIEmbeddings(),
            metadatas=[dict(data=p) for p in pool],
        )

    def sample(self, n: int) -> List[str]:
        """Sample n items from the pool"""
        if n > len(self._available):
            raise ValueError("Not enough items in pool")
        samples = np.random.choice(self._available, size=n, replace=False)
        return samples

    def choose(self, x: str) -> None:
        """Choose a specific item from the pool"""
        if x not in self._available:
            raise ValueError("Item not in pool")
        self._selected.append(x)
        self._available.remove(x)

    def approx_sample(self, x: str, k: int) -> None:
        """Given an approximation of x, return k similar"""
        # want to select extra, then remove previously chosen
        _k = k + len(self._selected)
        docs = self._db.max_marginal_relevance_search(x, k=_k, fetch_k=5 * _k)
        docs = [d.metadata["data"] for d in docs]
        # remove previously chosen
        docs = [d for d in docs if d not in self._selected]
        # select k
        return docs[:k]

    def reset(self) -> None:
        """Reset the pool"""
        self._selected = []
        self._available = self._pool[:]

    def __len__(self) -> int:
        return len(self._pool)

    def __repr__(self) -> str:
        return f"Pool of {len(self)} items with {len(self._selected)} selected"

    def __str__(self) -> str:
        return f"Pool of {len(self)} items with {len(self._selected)} selected"

    def __iter__(self):
        return iter(self._available)



from langchain.tools import BaseTool
from .asktell import AskTellFewShotTopk
from .pool import Pool
from typing import *
import os
import pandas as pd


class BOLiftTool(BaseTool):
    name = "Experiment Designer"
    description = "Propose or predict experiments using stateful ask-and-tell. "
    "Syntax: Tell {CSV_FILE}. Adds training examples to model. "
    "Ask. Returns best experiment to run next. "
    asktell: AskTellFewShotTopk = None
    pool: Pool = None

    def __init__(self, pool: Pool, asktell: Optional[AskTellFewShotTopk] = None, ):
        # call the parent class constructor
        super(BOLiftTool, self).__init__()

        if asktell is None:
            asktell = AskTellFewShotTopk()
        self.asktell = asktell
        self.pool = pool

    def _run(self, query: str) -> str:
        if query.startswith("Ask"):
            cmd = "ask"
        elif query.startswith("Tell"):
            cmd = "tell"
            arg = query[4:].strip()
        else:
            return "Invalid command. Syntax: Ask. Returns best experiment to run next. Tell {CSV_FILE}. Adds training examples to model."
        if cmd == "ask":
            results = self.asktell.ask(self.pool)
            return f"Best experiments to run next: {results[0][0]}"
        elif cmd == "tell":
            # check the path exists
            if not os.path.exists(arg):
                return f"File {arg} does not exist."
            # load the data column 0 is x, column 1 is y
            data = pd.read_csv(arg, header=None)
            for i in range(len(data)):
                self.asktell.tell(data.iloc[i, 0], data.iloc[i, 1])
            return f"Added {len(data)} training examples."
            
    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()



from .version import __version__
from .asktell import AskTellFewShotMulti, AskTellFewShotTopk
from .asktellfinetuning import AskTellFinetuning

try:
    from .asktellGPR import AskTellGPR
except ImportError:
    print("GPR Packages not installed. Do `pip install bolift[gpr]` to install them")
from .asktellRidgeRegression import AskTellRidgeKernelRegression
from .asktellNearestNeighbor import AskTellNearestNeighbor
from .pool import Pool
from .tool import BOLiftTool



import numpy as np
from .asktellGPR import AskTellGPR
from .llm_model import GaussDist


class AskTellRidgeKernelRegression(AskTellGPR):
    def __init__(self, alpha=1, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha

    def _set_regressor(self):
        self.coefficients = None
        self.train_x = None
        self.train_y = None
        self.mean_x = None
        self.std_x = None
        self.mean_y = None
        self.std_y = None

    def _dot_product_kernel(self, X1, X2):
        return X1.dot(X2.T)

    def _normalize(self, X, mean, std):
        if mean is None or std is None:
            raise ValueError(
                "Mean and standard deviation must be set for normalization"
            )
        return (X - mean) / std

    def _train(self, X, y):
        self.train_x = np.array(self._query_cache(X), dtype=np.float64)
        self.mean_x = np.mean(self.train_x, axis=0)
        self.std_x = np.std(self.train_x, axis=0)

        self.train_y = np.array(y, dtype=np.float64)
        self.mean_y = np.mean(self.train_y, axis=0)
        self.std_y = np.std(self.train_y, axis=0)

        self.train_x = self._normalize(self.train_x, self.mean_x, self.std_x)
        self.train_y = self._normalize(self.train_y, self.mean_y, self.std_y)

        kernel_train = self._dot_product_kernel(self.train_x, self.train_x)
        ridge_matrix = self.alpha * np.eye(kernel_train.shape[0])

        # Add a small diagonal matrix to the kernel matrix to ensure it's PSD
        eps = 1e-8
        kernel_train += eps * np.eye(kernel_train.shape[0])

        self.coefficients = np.linalg.solve(kernel_train + ridge_matrix, self.train_y)

    def _predict(self, X):
        if len(X) == 0:
            raise ValueError("X is empty")
        embedding = np.array(self._query_cache(X), dtype=np.float64)

        # Normalize the input features
        embedding = self._normalize(embedding, self.mean_x, self.std_x)

        kernel_test = self._dot_product_kernel(embedding, self.train_x)
        mean = np.dot(kernel_test, self.coefficients) * self.std_y + self.mean_y
        results = [GaussDist(m, s) for m, s in zip(mean, np.zeros_like(mean))]
        return results, 0



from setuptools import setup

exec(open("bolift/version.py").read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="bolift",
    version=__version__,
    description="BayesOPT with LIFT",
    author="Andrew White",
    author_email="andrew.white@rochester.edu",
    url="https://github.com/whitead/bolift",
    license="MIT",
    packages=["bolift"],
    install_requires=[
        "numpy",
        "langchain",
        "openai",
        "faiss-cpu",
        "scipy",
        "pandas",
        "tiktoken",
    ],
    extras_require={"gpr": ["scikit-learn", "torch", "botorch", "gpytorch"]},
    test_suite="tests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
